%% Ein einfaches inverses Problem.
%% Zuerst: die Vorwärts-Abbildung
% Gegeben sind Originaldaten; hier: Punkt-Koordinaten.

p =[2; 5];

%% 
% Gegeben ist weiters eine Abbildungsfunktion; hier: eine Matrix. Sie transformiert 
% Originaldaten -> Bilddaten. 

% das ist eine "gutmütige" Abbildung
% A=[1 3 
%    2 4];

%für diese Matrix wird es dann nicht mehr so einfach...
   A=[1 2 
      2 4.1];

%% 
% Das sind die Bildpunkt-Koordinaten

 q = A*p
%% und jetzt umgekehrt: Inverses Problem
% gegeben q, was war der Ausgangswert p?  Wenn q exakt berechnet oder gemessen 
% wurde, lässt sich die Abbildung problemlos umkehren.

 A\q % Umkehrabbildung
%% 
% Jetzt angenommen, es liegen viele Messwerte für q vor, aber diese q-Daten 
% sind ein alle ein bisserl verrauscht. 

% Das ist Datensatz mit 100 leicht gestörten q-Daten
q_gestoert = q + 0.1*randn(2,100);

% wodurch die rekonstruierten p-Werte auch gestört sind
p_gestoert = A\q_gestoert;

% Ein Plot der gestörten Messdaten und der daraus rekonstruierten
% Originalwerte: Für die erste Matrix gilt: 
% kleine Störung der Messwerte -> kleine Unterschiede bei Rekonstruktion
%
% Für zweite Matrix: 
% kleine Störung -> riesige Unterschiede ruinieren Rekonstruktion
plot(q_gestoert(1,:),q_gestoert(2,:),'or',p_gestoert(1,:),p_gestoert(2,:),'ob')
axis equal
title({'rot: Bildpunkte, gestört','Blau: daraus rekonstruierte Ausgangswerte'})
%% regularisierte Lösung:
% löse das System der Normalengleichungen mit ein klein wenig Einheitsmatrix 
% dazu: (A^T A + alpha I) p = A^T q 
% 
% Ergibt deutlich stabilere Rekonstruktion!

alpha=0.01; % Regularisierungsparameter. Wahl nicht ganz einfach.
p_reg = (A'*A+alpha*eye(2))\A'*q_gestoert;
plot(q_gestoert(1,:),q_gestoert(2,:),'or',p_gestoert(1,:),p_gestoert(2,:),'ob',p_reg(1,:),p_reg(2,:),'og')
title({'rot: Bildpunkte, gestört','Blau: rekonstruierte Ausgangswerte','Grün: regularisierte Ausgangswerte'})
grid on
axis equal
%% regularisierte Lösung, Rechenweg über SVD
% für tieferes Verständnis und weitereführende Arbeit ist SVD notwendig. Als 
% Kochrezept ist der Rechenweg über Normalengleichungen vielleicht einfacher. 

[U, S, V]=svd(A);
Sreg=zeros(size(S));
Sreg=diag(diag(S)./(diag(S).^2 + alpha));
p_regSVD = V*Sreg*U'*q_gestoert;

%Ergebnisse bis auf kleine Rundungsfehler ident.
norm(p_reg-p_regSVD)